import inspect
import pickle
from datetime import timedelta
from typing import Optional, Union, Any
from discord.ext import commands

from kasushi.exceptions import InvalidConfigurationError, InvalidRequirementsError

try:
    import aiomcache
except ImportError:
    raise InvalidRequirementsError("aiomcache is not installed.")


class Cache:
    def __init__(self, bot: commands.Bot, raw_cache: aiomcache.Client):
        self.bot = bot
        self._raw_cache = raw_cache

    async def bget(self, key: bytes, *, default: Optional[bytes] = None) -> bytes:
        return await self._raw_cache.get(key, default=default)

    async def bset(self, key: bytes, value: bytes, *, expire: Optional[timedelta] = None) -> bool:
        """
        Set a key in the cache. If expire is not None, the key will expire after the given time.
        """
        if expire is None:
            exptime = 0
        else:
            exptime = expire.total_seconds()

        return await self._raw_cache.set(key, value, exptime=exptime)

    async def bdelete(self, key: bytes) -> bool:
        """
        Delete a key from the cache.

        Returns True if the key was deleted, False if it was not found.
        """
        return await self._raw_cache.delete(key)

    def encode(self, obj: Any) -> bytes:
        return pickle.dumps(obj)

    def decode(self, data: bytes) -> Any:
        return pickle.loads(data)

    async def get(self, key: str, *, default: Any = None) -> str:
        return self.decode(await self.bget(key.encode(), default=self.encode(default)))

    async def set(self, key: str, value: Any, *, expire: Optional[timedelta] = None) -> bool:
        return await self.bset(key.encode(), self.encode(value), expire=expire)

    async def delete(self, key: str) -> bool:
        return await self.bdelete(key.encode())


class CacheCog(commands.Cog):
    def __init__(self, bot: commands.Bot) -> None:
        self.bot = bot
        self.raw_cache: Optional[aiomcache.Client] = None
        self.cache: Optional[Cache] = None

    async def config_check(self) -> dict:
        config: dict = self.bot._kasushi_config
        cache_config: Optional[dict] = config.get('cache')
        if not cache_config:
            raise InvalidConfigurationError(
                message="Cache isn't configured. Please set the `cache` in your config."
            )

        cache_server = cache_config.get('server_ip')
        if not cache_server:
            raise InvalidConfigurationError(
                message="Cache isn't configured. Please set the `server_ip` key in your cache config."
            )

        server_port = cache_config.setdefault('server_port', 11211)

        return config

    async def cog_load(self) -> None:
        config = await self.config_check()
        mc = aiomcache.Client(config["cache"]["server_ip"], config["cache"]["server_port"])
        self.raw_cache = mc
        self.cache = Cache(self.bot, mc)
        self.bot.cache = self.cache

    async def cog_unload(self) -> None:
        pass


async def async_setup(bot: commands.Bot):
    """
    The async setup function defining the jishaku.cog and jishaku extensions.
    """

    await bot.add_cog(CacheCog(bot=bot))


def setup(bot: commands.Bot):
    """
    The setup function defining the jishaku.cog and jishaku extensions.
    """

    if inspect.iscoroutinefunction(bot.add_cog):
        return async_setup(bot)

    bot.add_cog(CacheCog(bot=bot))



class KasushiError(Exception):
    """Base class for exceptions in this module."""
    pass


class InvalidConfigurationError(KasushiError):
    """
    Exception raised for errors in the configuration.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class InvalidRequirementsError(KasushiError):
    """
    Exception raised for errors in the requirements.
    It typically means you need to install a missing package.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message



import inspect
from typing import Optional

from discord.ext import commands

from kasushi.exceptions import InvalidConfigurationError
from kasushi.ipc.client import IPCClient
from kasushi.ipc.server import IPCServer


class IPCCog(commands.Cog):
    def __init__(self, bot: commands.AutoShardedBot) -> None:
        self.server = None
        self.bot = bot

    async def config_check(self) -> dict:
        config: dict = self.bot._kasushi_config['ipc']

        shared_secret = config.get('shared_secret')
        if not shared_secret:
            raise InvalidConfigurationError('IPC `shared_secret` is not set')
        elif shared_secret == '' or shared_secret == 'secret':
            raise InvalidConfigurationError('IPC `shared_secret` is not secret enough')

        try:
            ipc_server = config['ipc_server']
        except KeyError:
            raise InvalidConfigurationError('IPC `ipc_server` is not set')
        else:
            try:
                ipc_server_host = ipc_server['host']
            except KeyError:
                raise InvalidConfigurationError('IPC `ipc_server.host` is not set')

            try:
                ipc_server_port = ipc_server['port']
            except KeyError:
                raise InvalidConfigurationError('IPC `ipc_server.port` is not set')

        try:
            ipc_client = config['client']
        except KeyError:
            raise InvalidConfigurationError('IPC `client` is not set')
        else:
            try:
                ipc_client_server_url = ipc_client['server_url']
            except KeyError:
                raise InvalidConfigurationError('IPC `client.server_url` is not set')
            else:
                if not ipc_client_server_url.startswith('http'):
                    raise InvalidConfigurationError('IPC `client.server_url` must start with http or https')

        config.setdefault('handlers', [])

        return config

    async def cog_load(self) -> None:
        config = await self.config_check()

    async def on_ready(self) -> None:
        config = await self.config_check()

        shard_zero = self.bot.get_shard(0)
        if shard_zero:
            self.server = IPCServer(config['server'])

            for handler in config['handlers']:
                self.server.add_handler(handler)

            await self.server.async_setup()

        self.bot.ipc = IPCClient(self.bot, config['client'])
        for handler in config['handlers']:
            self.bot.ipc.add_handler(handler)

        await self.bot.ipc.async_setup()

    async def cog_unload(self) -> None:
        if self.server:
            await self.server.async_teardown()

        await self.bot.ipc.async_teardown()


async def async_setup(bot: commands.AutoShardedBot):
    """
    The async setup function defining the jishaku.cog and jishaku extensions.
    """

    await bot.add_cog(IPCCog(bot=bot))


def setup(bot: commands.AutoShardedBot):
    """
    The setup function defining the jishaku.cog and jishaku extensions.
    """

    if inspect.iscoroutinefunction(bot.add_cog):
        return async_setup(bot)

    bot.add_cog(IPCCog(bot=bot))



import discord
import pytest
import asyncio

from kasushi.ipc.handlers import GuildInfoHandler
from kasushi.ipc.client import IPCClient
from kasushi.ipc.server import IPCServer
from utils import setup_logger

setup_logger()

configuration = \
    {
        "shared_secret": "secret",  # This is used to authenticate with the IPC server.
        "server": {
            "host": "0.0.0.0",
            "port": 12321,
        },
        "client": {
            "server_url": "http://127.0.0.1:12321",
        },
        "handlers": [GuildInfoHandler],  # See above for handlers.
    }


class BotMock:
    def __init__(self):
        self.shards = {}
        self.guilds = []


@pytest.mark.asyncio
async def test_simple_login():
    botMockClient1 = BotMock()
    botMockClient2 = BotMock()

    botMockClient1.shards = {0: None}
    botMockClient1.guilds = [discord.Object(id=0), discord.Object(id=1), discord.Object(id=2)]

    botMockClient2.shards = {1: None}
    botMockClient2.guilds = [discord.Object(id=10), discord.Object(id=11)]

    server = IPCServer(configuration)
    await server.async_setup()

    client1 = IPCClient(botMockClient1, configuration)
    client2 = IPCClient(botMockClient2, configuration)
    await client1.async_setup()
    await client2.async_setup()

    await asyncio.sleep(2)
    assert len(server._active_ws) == 2
    assert 0 in server.shard_to_ws_mapping.keys()
    assert 1 in server.shard_to_ws_mapping.keys()

    await server.async_teardown()
    await client1.async_teardown()
    await client2.async_teardown()



import logging
from typing import Dict, Any, List, Optional, Type

import aiohttp.web
from aiohttp.web_ws import WebSocketResponse
from discord.ext import commands

from .base import Handler, LoginHandler

logger = logging.getLogger(__name__)


class IPCServer:
    def __init__(self, config: dict, *args, **kwargs):
        self.config = config

        self._active_ws: List[WebSocketResponse] = []

        self.return_paths: Dict[str, WebSocketResponse] = {}

        self.handlers: Dict[str, Handler] = {}

        self.shard_to_ws_mapping: Dict[int, WebSocketResponse] = {}
        self.guild_to_ws_mapping: Dict[int, WebSocketResponse] = {}
        self.add_handler(LoginHandler)

    def add_handler(self, handler: Type[Handler]):
        self.handlers[handler.name] = handler(self)

    async def index(self, request):
        return aiohttp.web.Response(text='This is a websocket IPC server for Kasushi IPC')

    async def websocket_handler(self, request):
        # Got websocket connection
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)
        self._active_ws.append(ws)
        logged_in = False

        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                json_data = msg.json()

                if logged_in:
                    await self.handle_incoming_message(ws, json_data)
                else:
                    await self.handle_login_message(ws, json_data)

        self._active_ws.remove(ws)
        return ws

    async def handle_incoming_message(self, ws, data):
        logger.debug(f'[{ws}] {data}')
        type = data.get('type')
        handler = data.get('handler')
        rtoken = data.get('rtoken')
        if type == 'request':
            handler_class = self.handlers.get(handler)
            if handler_class:
                if rtoken and not handler_class.server_reply:
                    self.return_paths[rtoken] = ws
                await handler_class.server_dispatch(ws, data)
            else:
                logger.warning(f'[{ws}] Handler {handler} not found, cannot forward')
        else:
            if rtoken:
                try:
                    await self.return_paths[rtoken].send_json(data)
                except KeyError:
                    logger.warning(f'[{ws}] No return path for {rtoken}')
                    return

    async def handle_login_message(self, ws, data):
        data['handler'] = 'login'
        data['type'] = 'request'
        await self.handle_incoming_message(ws, data)

    async def async_setup(self):
        app = aiohttp.web.Application()
        app.add_routes([
            aiohttp.web.get('/', self.index),
            aiohttp.web.get('/ws', self.websocket_handler),
        ])
        self._app = app
        self._runner = aiohttp.web.AppRunner(app)
        await self._runner.setup()
        self._site = aiohttp.web.TCPSite(self._runner, self.config.get('server_listen_host', '0.0.0.0'),
                                         self.config.get('server_listen_port', 12321))
        await self._site.start()

    async def async_teardown(self):
        logger.debug("IPC Server closing")
        for ws in self._active_ws[:]:
            logger.debug(f"Closing {ws}")
            await ws.close()
            logger.debug(f"Closed {ws}")

        await self._site.stop()
        logger.debug("IPC Server closed")



import asyncio
import json
import logging
import random
import uuid
from typing import Dict, Type

import aiohttp
from discord.ext import commands

from .base import Handler, LoginHandler

logger = logging.getLogger(__name__)


class WaitingEvent(asyncio.Event):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response = None


class IPCClient:
    type: str = None

    def __init__(self, bot: commands.AutoShardedBot, config: dict):
        self.handlers: Dict[str, Handler] = {}
        self.bot = bot
        self.config = config
        self._URL = config['client']['server_url'] + "/ws"
        self.online = False
        self.closed = False
        self.ws = None
        self.session = None
        self.waiting_events = {}

        self.add_handler(LoginHandler)

    def add_handler(self, handler: Type[Handler]):
        self.handlers[handler.name] = handler(self)

    async def handle_message(self, data):
        type = data.get('type')
        rtoken = data.get('rtoken')
        handler = data.get('handler')
        if rtoken and type == 'response':
            event = self.waiting_events.get(rtoken)
            if event:
                event.response = data['data']
                event.set()
                return

        ret = await self.handlers[handler].get_response(data['data'])
        if ret:
            ret['handler'] = handler
            ret['rtoken'] = rtoken
            ret['type'] = 'response'
            await self.send(ret)

    async def async_setup(self):
        self.session = aiohttp.ClientSession()
        asyncio.create_task(self.background())

    async def background(self):
        first = True
        logger.debug("Starting IPC Client")
        while not self.closed:
            if first:
                logger.debug("Starting first IPC Client connection to server")
                first = False
            else:
                logger.debug("Starting IPC Client re-connection to server in 5 seconds")
                await asyncio.sleep(2)

            async with self.session.ws_connect(self._URL) as ws:
                logger.debug("IPC Client websocket connection established")
                self.ws = ws
                await self.handlers['login'].send_request()
                logger.debug("IPC Client login sent")

                async for msg in self.ws:
                    logger.debug("IPC Client message recv'd: " + str(msg))
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        await self.handle_message(json.loads(msg.data))
                    elif msg.type in (aiohttp.WSMsgType.CLOSED,
                                      aiohttp.WSMsgType.ERROR):
                        break
            logger.debug("IPC Client websocket closed")
            await asyncio.sleep(2)

    async def async_teardown(self):
        logger.debug("IPC Client closing")
        self.closed = True
        self.online = False

        if self.ws:
            logger.debug("IPC Client closing websocket connection")
            await self.ws.close()
            logger.debug("IPC Client websocket connection closed")

        if self.session:
            logger.debug("IPC Client closing session")
            await self.session.close()
            logger.debug("IPC Client session closed")

    async def send_wait(self, data):
        data['rtoken'] = uuid.uuid4().hex
        event = WaitingEvent()
        self.waiting_events[data['rtoken']] = event
        await self.send(data)
        await event.wait()
        del self.waiting_events[data['rtoken']]

        return event.response

    async def send(self, data):
        await self.ws.send_json(data)



from discord.ext import commands

import logging
import logging.config

logger = logging.getLogger(__name__)


async def configure(bot: commands.Bot, settings: dict):
    bot._kasushi_settings = settings


DEFAULT_LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s - %(message)s'
        },
        'complete': {
            'format': '%(asctime)s - %(levelname)s - %(filename)s - %(message)s',
        },
    },
    'handlers': {
        'default': {
            'level': 'DEBUG',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': True
        },
    }
}


def setup_logger():
    logging.config.dictConfig(DEFAULT_LOGGING)
    logging.info('Hello, log')



from typing import Union, Optional, TYPE_CHECKING

from aiohttp.web_ws import WebSocketResponse

if TYPE_CHECKING:
    from .client import IPCClient
    from .server import IPCServer


class Handler:
    server_reply = False
    name: str = None
    wait = True

    def __init__(self, ipc: Union['IPCClient', 'IPCServer']):
        self.ipc = ipc

    async def send_request(self, *args, **kwargs):
        self.ipc: 'IPCClient'
        data = {'handler': self.name, 'type': 'request', 'data': await self.get_request_data(*args, **kwargs)}

        if self.wait:
            await self.ipc.send_wait(data)
        else:
            await self.ipc.send(data)

    async def get_request_data(self, *args, **kwargs):
        self.ipc: 'IPCClient'
        return {}

    async def server_dispatch(self, sender_ws: WebSocketResponse, data: dict):
        self.ipc: 'IPCServer'
        pass

    async def get_response(self, data: dict) -> Optional[dict]:
        self.ipc: 'IPCClient'
        pass


class LoginHandler(Handler):
    server_reply = True
    name = 'login'

    async def get_request_data(self):
        self.ipc: 'IPCClient'
        return {"shared_secret": self.ipc.config['shared_secret'],
                'guilds': [g.id for g in self.ipc.bot.guilds],
                'shards': list(self.ipc.bot.shards.keys())}

    async def server_dispatch(self, sender_ws: WebSocketResponse, data: dict):
        self.ipc: 'IPCServer'
        ret = {
            'handler': data['handler'],
            'type': 'response',
            'rtoken': data['rtoken'],
            'data': await self.get_response(data, sender_ws)
        }
        await sender_ws.send_json(ret)

    async def get_response(self, data: dict, sender_ws: WebSocketResponse) -> Optional[dict]:
        self.ipc: 'IPCServer'  # Only in this case
        if data['data']['shared_secret'] == self.ipc.config['shared_secret']:
            guilds = data['data']['guilds']
            shards = data['data']['shards']

            for guild_id in guilds:
                self.ipc.guild_to_ws_mapping[guild_id] = sender_ws

            for shard_id in shards:
                self.ipc.shard_to_ws_mapping[shard_id] = sender_ws

            return {"success": True, 'message': 'You are now authenticated'}
        else:
            return {"success": False, 'message': 'Invalid secret'}




from typing import Optional

from aiohttp.web_ws import WebSocketResponse

from .base import Handler
from .client import IPCClient
from .server import IPCServer


class GuildInfoHandler(Handler):
    name = 'guild_info'

    async def get_request_data(self, guild_id: int):
        self.ipc: 'IPCClient'
        return {"guild_id": guild_id}

    async def server_dispatch(self, sender_ws: WebSocketResponse, data: dict):
        self.ipc: 'IPCServer'
        await self.ipc.guild_to_ws_mapping[data['data']['guild_id']].send_json(data)

    async def get_response(self, data: dict) -> Optional[dict]:
        self.ipc: 'IPCClient'
        guild = self.ipc.bot.get_guild(data['guild_id'])
        if guild:
            return {
                'success': True,
                'guild': {
                    'name': guild.name,
                    'id': guild.id,
                    'member_count': guild.member_count,
                }
            }
        else:
            return {
                'success': False,
                'message': 'Guild not found'
            }
